from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

from snitch.exceptions import HandlerError
from snitch.helpers import (
    extract_actor_trigger_target,
    get_notification_model,
    send_event_to_user,
)

if TYPE_CHECKING:
    from snitch.backends import AbstractBackend
    from snitch.models import Event, Notification


class EventHandler:
    """Base event backend to generic even types."""

    should_notify: bool = True
    should_send: bool = True
    ephemeral: bool = False
    dispatch_config: Dict = {"args": ("actor", "trigger", "target")}
    action_type: Optional[str] = None
    action_id: Optional[str] = None
    title: str = ""
    text: str = ""
    delay: int = 0
    notification_backends: List[Type["AbstractBackend"]] = []

    template_email_async: bool = False

    @classmethod
    def extract_actor_trigger_target(cls, method: str, *args, **kwargs):
        """Extracts actor, trigger and target from the args and kwargs
        given as parameters. Override to implement a specific extractor.
        """
        if not isinstance(cls.dispatch_config, dict) or (
            "args" not in cls.dispatch_config and "kwargs" not in cls.dispatch_config
        ):
            raise HandlerError(_("The dispatch config is incorrect."))
        return extract_actor_trigger_target(
            config=cls.dispatch_config, args=args, kwargs=kwargs
        )

    def __init__(self, event: "Event", notification: "Notification" = None):
        self.event = event
        self.notification = notification

    def _default_dynamic_text(self) -> str:
        """Makes an event human readable."""
        text = "{} {}".format(str(self.event.actor), self.event.verb)
        if self.event.trigger:
            text = "{} {}".format(text, str(self.event.trigger))
        if self.event.target:
            text = "{} {}".format(text, str(self.event.target))
        return text

    def get_text(self) -> str:
        """Override to handle different human readable implementations."""
        return self.text or self._default_dynamic_text()

    def get_title(self) -> str:
        """Gets the title for the event. To be hooked."""
        return self.title

    def get_action_type(self) -> Optional[str]:
        """Gets the action type depending on the verb. To be hooked."""
        return self.action_type

    def get_action_id(self) -> Optional[str]:
        """Gets the action depending on the verb. To be hooked."""
        return self.action_id

    def get_delay(self) -> int:
        """Returns and in, number of seconds, that corresponds with the time
        the notification should be delayed.
        """
        return self.delay

    def get_language(self, user=None) -> str:
        """Gets the locale for the given used. By default, users the LANGUAGE_CODE
        value from settings.
        """
        return settings.LANGUAGE_CODE

    def get_extra_data(self) -> Dict:
        """Adds extra meta data to the backend."""
        return {}

    def audience(self) -> QuerySet:
        """Gets the audience of the event. None by default, to be hooked by the user."""
        User = get_user_model()
        return User.objects.none()

    def notify(self):
        """If the event is not ephemeral, creates a notification fot each user in the
        audience. In other case, only sends the notification, but doesn't save
        into the database.
        """
        if not self.ephemeral:
            # Creates a notification
            Notification = get_notification_model()
            for user in self.audience():
                notification = Notification(event=self.event, user=user)
                notification.save()
        else:
            # Only sends the event to the user
            for user in self.audience():
                send_event_to_user(event=self.event, user=user)


class EventManager:
    """The event manager in the responsible of handling the registration of the
    handlers with the verbs.
    """

    def __init__(self):
        self._registry: Dict[str, Type["EventHandler"]] = {}
        self._verbs: Dict = {}

    def register(
        self, verb: str, handler: Type["EventHandler"], verbose: Optional[str] = None
    ):
        """Register a handler with a verb, and the verbose form of the verb."""
        if not issubclass(handler, EventHandler):
            raise HandlerError(
                _(f"The handler {handler} have to inherit from EventHandler.")
            )
        self._verbs[verb] = verbose if verbose else verb
        self._registry[verb] = handler

    def choices(self) -> Tuple:
        """Gets a tuple of tuples with the registers verbs and its verbose form, to be
        used as choices."""
        return tuple(self._verbs.items())

    def handler_class(self, verb) -> Type[EventHandler]:
        """Returns a class instance of the handler for the given verb."""
        return self._registry.get(verb, EventHandler)

    def handler(
        self, event: "Event", notification: "Notification" = None
    ) -> EventHandler:
        """Returns an instance of the handler for the given event."""
        return self.handler_class(event.verb)(event, notification=notification)


# This global object represents the singleton event manager object
manager: EventManager = EventManager()



from typing import List, Optional, Union

from celery.task import task
from django.core.mail import EmailMultiAlternatives

from snitch.helpers import get_notification_model


@task(serializer="json")
def send_notification_task(notification_pk: int) -> Optional[bool]:
    """A Celery task to send push notifications related with a given Notification
    model."""

    Notification = get_notification_model()

    try:
        notification = Notification.objects.get(pk=notification_pk)
    except Notification.DoesNotExist:
        return False
    notification.send(send_async=False)
    return None


@task(serializer="json")
def send_email_asynchronously(
    subject: str, message_txt: str, message: str, from_email: str, to: Union[List, str]
):
    """Sends an email as a asynchronous task."""
    email = EmailMultiAlternatives(
        subject=subject, body=message_txt, from_email=from_email, to=to
    )
    email.attach_alternative(message, "text/html")
    email.send()
    return True



from typing import TYPE_CHECKING, Dict

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import translation
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel

from snitch.handlers import manager

if TYPE_CHECKING:
    from snitch import EventHandler
    from snitch.backends import AbstractBackend

User = get_user_model()


class EventType(models.Model):
    """Explicit model for Event types, represented by the event verb. It's used to
    enable or disable the generation of notifications.
    """

    verb = models.CharField(
        max_length=255, null=True, choices=manager.choices(), unique=True
    )
    enabled = models.BooleanField(default=True, verbose_name=_("enabled"))

    class Meta:
        verbose_name = _("event type")
        verbose_name_plural = _("event types")
        ordering = ("verb",)

    def __str__(self) -> str:
        return self.verb


class Event(TimeStampedModel):
    """A 'event' is generated when an 'actor' performs 'verb', involving 'action',
    in the 'target'.

     It could be:
        <actor> <verb>
        <actor> <verb> <target>
        <actor> <verb> <trigger> <target>

    Reference: http://activitystrea.ms/specs/atom/1.0/
    """

    actor_content_type = models.ForeignKey(
        ContentType,
        related_name="actor_actions",
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("actor content type"),
    )
    actor_object_id = models.PositiveIntegerField(_("actor object id"), null=True)
    actor = GenericForeignKey("actor_content_type", "actor_object_id")

    verb = models.CharField(
        _("verb"), max_length=255, null=True, choices=manager.choices()
    )

    trigger_content_type = models.ForeignKey(
        ContentType,
        related_name="trigger_actions",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("trigger content type"),
    )
    trigger_object_id = models.PositiveIntegerField(
        _("trigger object id"), blank=True, null=True
    )
    trigger = GenericForeignKey("trigger_content_type", "trigger_object_id")

    target_content_type = models.ForeignKey(
        ContentType,
        related_name="target_actions",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        verbose_name=_("target content type"),
    )
    target_object_id = models.PositiveIntegerField(
        _("target object id"), blank=True, null=True
    )
    target = GenericForeignKey("target_content_type", "target_object_id")

    notified = models.BooleanField(_("notified"), default=False)

    class Meta:
        verbose_name = _("event")
        verbose_name_plural = _("events")

    def __str__(self) -> str:
        handler = self.handler()
        return handler.get_text()

    def handler(self, notification: "Notification" = None) -> "EventHandler":
        """Gets the handler for the event."""
        return manager.handler(self, notification=notification)

    def notify(self) -> None:
        """Creates the notifications associated to this action, ."""
        handler = self.handler()
        if handler.should_notify:
            handler.notify()
            self.notified = True
            self.save()

    def save(self, *args, **kwargs) -> None:
        super().save(*args, **kwargs)
        if not self.notified:
            self.notify()


class AbstractNotification(TimeStampedModel):
    """A notification is sent to an user, and it's always related with an event."""

    event = models.ForeignKey(
        Event,
        verbose_name=_("event"),
        related_name="notifications",
        on_delete=models.CASCADE,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="notifications", on_delete=models.CASCADE
    )
    sent = models.BooleanField(_("sent"), default=False)
    received = models.BooleanField(_("received"), default=False)
    read = models.BooleanField(_("read"), default=False)

    class Meta:
        verbose_name = _("notification")
        verbose_name_plural = _("notifications")
        ordering = ("-created",)
        abstract = True

    def __str__(self) -> str:
        return f"'{str(self.event)}' to {str(self.user)}"

    def _task_kwargs(self, handler: "EventHandler") -> Dict:
        """Gets the kwargs for celery task, used in apply_async method."""
        kwargs = {}
        # Delay from event handler
        delay = handler.get_delay()
        if delay:
            kwargs["countdown"] = delay
        return kwargs

    def handler(self) -> "EventHandler":
        """Gets the handler for the notification."""
        return self.event.handler(notification=self)

    def send(self, send_async: bool = False) -> None:
        """Sends a push notification to the devices of the user."""
        from snitch.tasks import send_notification_task

        handler: "EventHandler" = self.event.handler()
        if handler.should_send:
            if send_async:
                send_notification_task.apply_async(
                    (self.pk,), **self._task_kwargs(handler)
                )
            else:
                # Activate language for translations
                if settings.USE_I18N:
                    language = handler.get_language(self.user)
                    translation.activate(language)
                for backend_class in handler.notification_backends:
                    backend: "AbstractBackend" = backend_class(self)
                    backend.send()
                self.sent = True
                self.save()

    def save(self, *args, **kwargs) -> None:
        """Overwrite to sending push notifications when saving."""
        is_insert: bool = self._state.adding
        super().save(*args, **kwargs)
        if is_insert:
            self.send()


class Notification(AbstractNotification):
    """Initial notification model that can be swappable."""

    class Meta(AbstractNotification.Meta):
        swappable: str = "SNITCH_NOTIFICATION_MODEL"
